//------------------------------------------------------------------------------
// CERN BE-CO-HT
// LHC Instability Trigger Distribution (LIST)
// https://ohwr.org/projects/list
//------------------------------------------------------------------------------
//
// unit name:   main
//
// description: A SystemVerilog testbench to exercise all the main features of
// LIST
//
//------------------------------------------------------------------------------
// Copyright CERN 2018
//------------------------------------------------------------------------------
// Copyright and related rights are licensed under the Solderpad Hardware
// License, Version 2.0 (the "License"); you may not use this file except
// in compliance with the License. You may obtain a copy of the License at
// http://solderpad.org/licenses/SHL-2.0.
// Unless required by applicable law or agreed to in writing, software,
// hardware and materials distributed under this License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
// or implied. See the License for the specific language governing permissions
// and limitations under the License.
//------------------------------------------------------------------------------

`include "mock_turtle_driver.svh"
`include "vhd_wishbone_master.svh"

`timescale 1ns/10fs

module main;

   reg rst_n = 0;
   reg clk_125m_pllref = 0;
   reg clk_20m_vcxo = 0;
   reg clk_ext = 0;

   wire a2btxp, a2btxn, b2atxp, b2atxn;

  always #50ns clk_ext <= ~clk_ext;
  always #4ns clk_125m_pllref <= ~clk_125m_pllref;
  always #20ns clk_20m_vcxo <= ~clk_20m_vcxo;

   initial begin
      repeat(20) @(posedge clk_20m_vcxo);
      rst_n = 1;
   end

   svec_list_top #
     (
      .g_simulation (1),
      .g_sim_bypass_vme (1),
      .g_dpram_initf("../../ip_cores/wr-cores/bin/wrpc/wrc_phy8_sim.bram")
      )
   DUT_A
     (
      .rst_n_i             (rst_n),
      .clk_125m_pllref_p_i (clk_125m_pllref),
      .clk_125m_pllref_n_i (~clk_125m_pllref),
      .clk_125m_gtp_p_i    (clk_125m_pllref),
      .clk_125m_gtp_n_i    (~clk_125m_pllref),
      .clk_20m_vcxo_i      (clk_20m_vcxo),
      .sim_wb_i            (HostA.out),
      .sim_wb_o            (HostA.in),
      .sfp_txp_o           (a2btxp),
      .sfp_txn_o           (a2btxn),
      .sfp_rxp_i           (b2atxp),
      .sfp_rxn_i           (b2atxn)
      );

   IVHDWishboneMaster HostA
     (
      .clk_i   (DUT_A.clk_sys_62m5),
      .rst_n_i (DUT_A.rst_sys_62m5_n));

   IMockTurtleIRQ IrqMonitorA (`MT_ATTACH_IRQ(DUT_A.cmp_mock_turtle));

   svec_list_top #
     (
      .g_simulation (1),
      .g_sim_bypass_vme (1),
      .g_dpram_initf("../../ip_cores/wr-cores/bin/wrpc/wrc_phy8_sim.bram")
      )
   DUT_B
     (
      .rst_n_i             (rst_n),
      .clk_125m_pllref_p_i (clk_125m_pllref),
      .clk_125m_pllref_n_i (~clk_125m_pllref),
      .clk_125m_gtp_p_i    (clk_125m_pllref),
      .clk_125m_gtp_n_i    (~clk_125m_pllref),
      .clk_20m_vcxo_i      (clk_20m_vcxo),
      .sim_wb_i            (HostB.out),
      .sim_wb_o            (HostB.in),
      .sfp_txp_o           (b2atxp),
      .sfp_txn_o           (b2atxn),
      .sfp_rxp_i           (a2btxp),
      .sfp_rxn_i           (a2btxn)
      );

   IVHDWishboneMaster HostB
     (
      .clk_i   (DUT_B.clk_sys_62m5),
      .rst_n_i (DUT_B.rst_sys_62m5_n));

   IMockTurtleIRQ IrqMonitorB (`MT_ATTACH_IRQ(DUT_B.cmp_mock_turtle));

   string fw = "../../../tests/firmware/rmq-udp/fw-rmq-udp.bin";

   const uint64_t mt_base = 'h2_0000;

   MockTurtleDriver drvA, drvB;

   initial begin

      @(posedge DUT_A.rst_sys_62m5_n);
      @(posedge DUT_A.clk_sys_62m5);

      drvA = new (HostA.get_accessor(), mt_base, IrqMonitorA, "DUTA");
      drvB = new (HostB.get_accessor(), mt_base, IrqMonitorB, "DUTB");

      drvA.init();
      drvB.init();

      drvA.enable_console_irq (0, 1);
      drvB.enable_console_irq (1, 1);

      drvA.load_firmware(0, fw, 1'b0);
      drvB.load_firmware(1, fw, 1'b0);

      drvA.reset_core(0, 0);
      drvB.reset_core(1, 0);

      forever begin
	 drvA.update ();
	 drvB.update ();
	 #1us;
      end

   end // initial begin

   /// ////////////////////////////////////////////////////////////////////////
   /// Silence Xilinx unisim DSP48A1 warnings about invalid OPMODE
   /// ////////////////////////////////////////////////////////////////////////
   initial begin
      force DUT_A.cmp_xwrc_board_svec.cmp_board_common.cmp_xwr_core.
	WRPC.LM32_CORE.gen_profile_medium_icache.U_Wrapped_LM32.cpu.
	  multiplier.D1.OPMODE_dly = 0;
      force DUT_A.cmp_xwrc_board_svec.cmp_board_common.cmp_xwr_core.
	WRPC.LM32_CORE.gen_profile_medium_icache.U_Wrapped_LM32.cpu.
	  multiplier.D2.OPMODE_dly = 0;
      force DUT_A.cmp_xwrc_board_svec.cmp_board_common.cmp_xwr_core.
	WRPC.LM32_CORE.gen_profile_medium_icache.U_Wrapped_LM32.cpu.
	  multiplier.D3.OPMODE_dly = 0;
      force DUT_B.cmp_xwrc_board_svec.cmp_board_common.cmp_xwr_core.
	WRPC.LM32_CORE.gen_profile_medium_icache.U_Wrapped_LM32.cpu.
	  multiplier.D1.OPMODE_dly = 0;
      force DUT_B.cmp_xwrc_board_svec.cmp_board_common.cmp_xwr_core.
	WRPC.LM32_CORE.gen_profile_medium_icache.U_Wrapped_LM32.cpu.
	  multiplier.D2.OPMODE_dly = 0;
      force DUT_B.cmp_xwrc_board_svec.cmp_board_common.cmp_xwr_core.
	WRPC.LM32_CORE.gen_profile_medium_icache.U_Wrapped_LM32.cpu.
	  multiplier.D3.OPMODE_dly = 0;
   end

endmodule // main
